/********************************************************************************
Copyright (C) 2013 Christophe Jeannin <chris.j84@free.fr>
Hugh Bailey <obs.jim@gmail.com>
Eric Bataille <e.c.p.bataille@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
********************************************************************************/

#include "twitchsw.h"
#include <iostream>
#include <sstream>
#include <rapidjson\rapidjson.h>
#include <rapidjson\stringbuffer.h>
#include <rapidjson\writer.h>
#include <rapidjson\document.h>
#include <rapidjson\prettywriter.h>
#include <cstdlib>
#include <cerrno>
#include <tchar.h>

#include <curl/curl.h>
#include <stdlib.h>



CURL *curl;
CURLcode res;

using namespace std;
using namespace rapidjson;

std::string buffer;

const char StaticJSON[] = "{\"channel\":{\"status\":\"\",\"game\":\"\"}}";
string channeltitle;
string gametitle;
bool titlesinorder = true;
//HWND hwndUpdateStatus;
rapidjson::Document document;
String lastSceneName;

TwitchSwitcher::TwitchSwitcher(HINSTANCE hinstMain, OBSHOTKEYPROC ToggleHotkeyProc)
{
	hinstDll = hinstMain;

	// The localization part
	pluginLocale = new LocaleStringLookup;

	if (!pluginLocale->LoadStringFile(TEXT("plugins/twitchsw/locale/en.txt")))
		AppWarning(TEXT("Could not open locale string file '%s'"), TEXT("plugins/twitchsw/locale/en.txt"));

	if (scmpi(API->GetLanguage(), TEXT("en")) != 0)
	{
		String pluginStringFile;
		pluginStringFile << TEXT("plugins/twitchsw/locale/") << API->GetLanguage() << TEXT(".txt");
		if (!pluginLocale->LoadStringFile(pluginStringFile))
			AppWarning(TEXT("Could not open locale string file '%s'"), pluginStringFile.Array());
	}

	timeToSleep = TTS_DEFAULT;
	thread = 0;
	timer = 0;
	bKillThread = true;
	
	altDoSwitch = 1;

	// Start with an empty list of window-scene assignments
	nWindowsDefined = 0;
	scenes.Clear();
	games.Clear();
	titles.Clear();

	config.Open(OBSGetPluginDataPath() + CONFIGFILENAME, true);
	ReadSettings();

	stopReq = CreateEvent(NULL, FALSE, FALSE, NULL);

	if (startAuto)
	{
		StartThread();
	}

	settings = new TwitchSwitcherSettings(this, ToggleHotkeyProc);
	OBSAddSettingsPane(settings);
}

TwitchSwitcher::~TwitchSwitcher()
{
	OBSRemoveSettingsPane(settings);
	delete settings;

	delete pluginLocale;

	StopThread();
	config.Close();
	CloseHandle(stopReq);
}


void TwitchSwitcher::ReadSettings()
{
	startAuto = config.GetInt(TEXT("General"), TEXT("StartAuto"));
	timeToSleep = (DWORD)config.GetInt(TEXT("General"), TEXT("CheckFrequency"));
	nWindowsDefined = config.GetInt(TEXT("GENERAL"), TEXT("nWindows"));
	altDoSwitch = config.GetInt(TEXT("GENERAL"), TEXT("AltDoSwitch"));
	oauthtoken = config.GetString(TEXT("GENERAL"), TEXT("OAuthToken"));
	mangame = config.GetString(TEXT("GENERAL"), TEXT("ManualGame"));
	mantitle = config.GetString(TEXT("GENERAL"), TEXT("ManualTitle"));
	//titlesinorder = config.GetInt(TEXT("General"), TEXT("TitlesInOrder"));
	toggleHotkey = (DWORD)config.GetInt(TEXT("GENERAL"), TEXT("ToggleHotkey"));
	toggleHotkey = (DWORD)config.GetInt(TEXT("GENERAL"), TEXT("ToggleHotkey"));
	scenes.Clear();
	games.Clear();
	titles.Clear();
	config.GetStringList(TEXT("SceneDef"), TEXT("title"), titles);
	config.GetStringList(TEXT("SceneDef"), TEXT("game"), games);
	config.GetStringList(TEXT("SceneDef"), TEXT("scene"), scenes);
	if ((int)games.Num() < nWindowsDefined) nWindowsDefined = games.Num();
	if ((int)scenes.Num() < nWindowsDefined) nWindowsDefined = scenes.Num();
	if ((int)titles.Num() < nWindowsDefined) nWindowsDefined = titles.Num();

	if (!timeToSleep)
		timeToSleep = TTS_DEFAULT;
	else
		TTS_CLAMP(timeToSleep);
}


void TwitchSwitcher::WriteSettings()
{
	config.SetString(TEXT("General"), TEXT("OAuthToken"), oauthtoken);
	config.SetString(TEXT("General"), TEXT("ManualTitle"), mantitle);
	config.SetString(TEXT("General"), TEXT("ManualGame"), mangame);
	//config.SetInt(TEXT("General"), TEXT("TitlesInOrder"), titlesinorder);
	config.SetInt(TEXT("General"), TEXT("StartAuto"), startAuto);
	config.SetInt(TEXT("General"), TEXT("CheckFrequency"), timeToSleep);
	config.SetInt(TEXT("General"), TEXT("nWindows"), nWindowsDefined);
	config.SetInt(TEXT("General"), TEXT("AltDoSwitch"), altDoSwitch);
	config.SetInt(TEXT("General"), TEXT("ToggleHotkey"), toggleHotkey);
	config.SetStringList(TEXT("SceneDef"), TEXT("title"), titles);
	config.SetStringList(TEXT("SceneDef"), TEXT("game"), games);
	config.SetStringList(TEXT("SceneDef"), TEXT("scene"), scenes);
}

std::string MBFromW(LPCWSTR pwsz) {
	int cch = WideCharToMultiByte(CP_ACP, 0, pwsz, -1, 0, 0, NULL, NULL);

	char* psz = new char[cch];

	WideCharToMultiByte(CP_ACP, 0, pwsz, -1, psz, cch, NULL, NULL);

	std::string st(psz);
	delete[] psz;

	return st;
}

std::wstring s2ws(const std::string& s)
{
	int len;
	int slength = (int)s.length() + 1;
	len = MultiByteToWideChar(CP_ACP, 0, s.c_str(), slength, 0, 0);
	wchar_t* buf = new wchar_t[len];
	MultiByteToWideChar(CP_ACP, 0, s.c_str(), slength, buf, len);
	std::wstring r(buf);
	delete[] buf;
	return r;
}

void MyLog(String str1, String str2="")
{
	std::wstring log = std::wstring(str1) + std::wstring(str2);
	Log((TCHAR*)log.c_str());
}

void MyLog(std::string str1, std::string str2="")
{
	std::wstring log = std::wstring(s2ws(str1)) + std::wstring(s2ws(str2));
	Log((TCHAR*)log.c_str());
}

DWORD TwitchSwitcher::Run()
{
	LARGE_INTEGER dueTime;

	if ((timer = CreateWaitableTimer(NULL, FALSE, NULL)) == NULL)
		return -1;
	dueTime.QuadPart = 0;
	SetWaitableTimer(timer, &dueTime, timeToSleep, NULL, NULL, FALSE);
	HANDLE handles[] = { timer, stopReq };

	String currentWindowText;
	while (!bKillThread)
	{
		String currentScene = OBSGetSceneName();
		if (currentScene != lastSceneName)
		{
			bool found = false;
			if (titlesinorder)
			{
				for (int i = 0; i < nWindowsDefined; i++)
				{
					String toCheck = GetScene(i);
					if (toCheck == currentScene && !found)
					{
						if (titlesinorder)
						{
							gametitle = MBFromW(GetGame(i));
							channeltitle = MBFromW(GetTitle(i));
							found = true;
							updatetwitch();
						}
					}

				}
			}
			if (!titlesinorder && scenes.HasValue(OBSGetSceneName()))
			{
				Log(L"Randomly picking a title");
				while (!found && nWindowsDefined > 0)
				{
					int random = rand() % nWindowsDefined;
					if (currentScene == scenes[random])
					{
						found = true;
						gametitle = MBFromW(GetGame(random));
						channeltitle = MBFromW(GetTitle(random));
						updatetwitch();
					}
				}
			}
			lastSceneName = currentScene;
		}

		WaitForMultipleObjects(2, handles, FALSE, INFINITE);
	}

	CloseHandle(timer);
	return 0;
}

void TwitchSwitcher::StartThread(HWND hDialog)
{
	if (!thread)
	{
		bKillThread = false;
		thread = CreateThread(NULL, 0, ThreadProc, (LPVOID)this, 0, NULL);

		if (hDialog && thread)
		{
			SetWindowText(GetDlgItem(hDialog, IDC_RUN), pluginLocale->LookupString(TEXT("Settings.Running")));
			SetWindowText(GetDlgItem(hDialog, IDC_STOP), pluginLocale->LookupString(TEXT("Settings.Stop")));
		}
	}
}


void TwitchSwitcher::StopThread(HWND hDialog)
{
	if (thread)
	{
		bKillThread = true;

		SetEvent(stopReq);
		WaitForSingleObject(thread, INFINITE);
		ResetEvent(stopReq);

		if (thread)
		{
			CloseHandle(thread);
			thread = 0;
		}
	}

	if (hDialog)
	{
		SetWindowText(GetDlgItem(hDialog, IDC_RUN), pluginLocale->LookupString(TEXT("Settings.NotRunning")));
		SetWindowText(GetDlgItem(hDialog, IDC_STOP), pluginLocale->LookupString(TEXT("Settings.Start")));
	}
}


void TwitchSwitcher::ApplyConfig(HWND hDialog)
{
	if (!hDialog) // Noes.. Dialog don't exist?
		return;

	// Get the listbox control handles
	HWND hwndMainScn = GetDlgItem(hDialog, IDC_MAINSCN);
	HWND hwndAltScn = GetDlgItem(hDialog, IDC_ALTSCN);
	HWND hwndAppList = GetDlgItem(hDialog, IDC_APPLIST);
	HWND hwndWSMap = GetDlgItem(hDialog, IDC_WSMAP);
	HWND hwndHotkey = GetDlgItem(hDialog, IDC_TOGGLEHOTKEY);
	HWND hwndOAuth = GetDlgItem(hDialog, IDC_FREQ2);
	HWND hwndManTitle = GetDlgItem(hDialog, IDC_FREQ3);
	HWND hwndManGame = GetDlgItem(hDialog, IDC_FREQ5);

	// Get the options names from the controls
	String newMainSceneName = GetCBText(hwndMainScn, CB_ERR);
	String newAltSceneName = GetCBText(hwndAltScn, CB_ERR);
	String newMainWndName = GetCBText(hwndAppList, CB_ERR);
	//String newoauthtoken = GetCBText(hwndOAuth, CB_ERR);
	String newoauthtoken = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";
	String newmantitle = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";
	String newmangame = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";
	GetDlgItemText(hDialog, IDC_FREQ2, newoauthtoken, 50);
	GetDlgItemText(hDialog, IDC_FREQ3, newmantitle, 50);
	GetDlgItemText(hDialog, IDC_FREQ5, newmangame, 50);

	// Get the new frequency from the frequency control
	DWORD newTimeToSleep = GetDlgItemInt(hDialog, IDC_FREQ, NULL, FALSE);
	TTS_CLAMP(newTimeToSleep);

	// Get the toggle hotkey if any
	toggleHotkey = (DWORD)SendMessage(hwndHotkey, HKM_GETHOTKEY, 0, 0);

	// Get the checkboxes state
	altDoSwitch = IsDlgButtonChecked(hDialog, IDC_ALTSWITCH);
	startAuto = IsDlgButtonChecked(hDialog, IDC_STARTAUTO);

	if (newAltSceneName.IsEmpty() && IsAltDoSwitch())
		StopThread(hDialog);	// if we're applying a bad config stop thread first

	if ((timeToSleep != newTimeToSleep) && thread && timer)
	{							// reset timer with new value
		LARGE_INTEGER dueTime;
		dueTime.QuadPart = 0;
		SetWaitableTimer(timer, &dueTime, newTimeToSleep, NULL, NULL, FALSE);
	}


	nWindowsDefined = SendMessage(hwndWSMap, LVM_GETITEMCOUNT, 0, 0);
	scenes.Clear();
	games.Clear();
	titles.Clear();
	for (int i = 0; i < nWindowsDefined; i++)
	{
		// Get the window name		
		String tit;
		tit.SetLength(256);
		ListView_GetItemText(hwndWSMap, i, 0, tit, 256);
		tit = String(tit.Array());
		titles.Add(tit);

		// Get the window name		
		String gme;
		gme.SetLength(256);
		ListView_GetItemText(hwndWSMap, i, 1, gme, 256);
		gme = String(gme.Array());
		games.Add(gme);

		// Get the scene name
		String scn;
		scn.SetLength(256);
		ListView_GetItemText(hwndWSMap, i, 2, scn, 256);
		scn = String(scn.Array());
		scenes.Add(scn);
	}

	// Set the new global values
	timeToSleep = newTimeToSleep;
	oauthtoken = newoauthtoken;
	mantitle = newmantitle;
	mangame = newmangame;
	//titlesinorder = IsDlgButtonChecked(hDialog, IDC_ORDER);
	//SendMessage(GetDlgItem(hDialog, IDC_ORDER), BM_SETCHECK, titlesinorder ? BST_CHECKED : BST_UNCHECKED, 0);

	WriteSettings(); // Write the settings to config
	updatetwitch();
}

size_t curl_write(void *ptr, size_t size, size_t nmemb, void *stream)
{
	buffer.append((char*)ptr, size*nmemb);
	return size*nmemb;
}

void TwitchSwitcher::Curl(std::string URL, String OAuthToken, bool Put, std::string PutData = "")
{
	curl_global_init(CURL_GLOBAL_ALL);
	curl = curl_easy_init();
	if (curl) {
		struct curl_slist *headers = NULL;

		string temp = "Authorization: OAuth " + MBFromW(OAuthToken);

		headers = curl_slist_append(headers, temp.c_str());
		headers = curl_slist_append(headers, "Content-Type: application/json");
		headers = curl_slist_append(headers, "Accept: application/json");
		headers = curl_slist_append(headers, "charsets: utf-8");
		res = curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
		curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0L);
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write);
		curl_easy_setopt(curl, CURLOPT_URL, URL.c_str());
		curl_easy_setopt(curl, CURLOPT_IPRESOLVE, CURL_IPRESOLVE_V4);
		curl_easy_setopt(curl, CURLOPT_DNS_USE_GLOBAL_CACHE, false);
		curl_easy_setopt(curl, CURLOPT_DNS_CACHE_TIMEOUT, 2);
		if (Put)
		{
			curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "PUT");
			curl_easy_setopt(curl, CURLOPT_POSTFIELDS, PutData.c_str());
		}
		res = curl_easy_perform(curl);
		if (res != CURLE_OK)
			fprintf(stderr, "curl_easy_perform() failed: %s\n",
			curl_easy_strerror(res));
		/* always cleanup */
		curl_easy_cleanup(curl);
	}

	//curl_global_cleanup();
}

void TwitchSwitcher::updatetwitch()
{
	Curl("https://api.twitch.tv/kraken/", oauthtoken, false);
	document.Parse(buffer.c_str());
	MyLog("Retrieved from Twitch: ", buffer);

	if (document["token"]["valid"].GetBool()) //Check that we Authed with token
	{
		Log(L"We authenticated correctly with OAuthToken");
		MyLog("Setting Stream Title: ", channeltitle);
		MyLog("Setting Game   Title: ", gametitle);
		buffer.clear();
		string channelurl = document["_links"]["channels"].GetString();
		Curl(channelurl, oauthtoken, false);

		StringBuffer sb1;
		PrettyWriter<StringBuffer> writer2(sb1);
		document.Accept(writer2);    // Accept() traverses the DOM and generates Handler events.

		document.Parse(StaticJSON);
		document["channel"]["status"].SetString(channeltitle.c_str(), document.GetAllocator());
		document["channel"]["game"].SetString(gametitle.c_str(), document.GetAllocator());

		StringBuffer sb;
		PrettyWriter<StringBuffer> writer(sb);
		document.Accept(writer);    // Accept() traverses the DOM and generates Handler events.
		
		Curl(channelurl, oauthtoken, true, sb.GetString());
		//
		//std::string status = "Stream Title: " + channeltitle + " Game Title: " + gametitle;

		//SetWindowTextA(hwndUpdateStatus, status.c_str());
		//mLabel->SetWindowText("test");
		for (int i = 0; i < 5; i++) //check if our update worked
		{
			buffer.clear();
			Curl(channelurl, oauthtoken, false);
			sb.Clear();
			writer.Reset(sb);
			document.Parse(buffer.c_str());
			if (document["status"].GetString() == channeltitle && document["game"].GetString() == gametitle)
			{
				Log(L"Channel updated correctly");
				//MyLog("Retrieved from Twitch: ", buffer);
				break;
			}
			else
			{
				Log(L"Channel hasn't updated yet, trying again in 2 seconds");
			}
			if (i == 4)
			{
				Log(L"Channel update failed, trying again");
				updatetwitch();
			}


			Sleep(2000);
		}

	}
	
	buffer.clear();
	std::getchar();
}

void TwitchSwitcher::VerifyUpdate()
{

}
