/********************************************************************************
 Copyright (C) 2013 Christophe Jeannin <chris.j84@free.fr>
                    Eric Bataille <e.c.p.bataille@gmail.com>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
********************************************************************************/

#include "twitchsw.h"

TwitchSwitcherSettings::TwitchSwitcherSettings(TwitchSwitcher *thePlugin, OBSHOTKEYPROC ToggleHotkeyProc):
                            SettingsPane(),
                            thePlugin(thePlugin),
                            ToggleHotkeyProc(ToggleHotkeyProc)
{
    pChange = false;
}

TwitchSwitcherSettings::~TwitchSwitcherSettings()
{
}

CTSTR TwitchSwitcherSettings::GetCategory() const
{
	return PluginStr("Plugin.SettingsName");
}

HWND TwitchSwitcherSettings::CreatePane(HWND parentHwnd)
{
    hwnd = CreateDialogParam(thePlugin->hinstDll, MAKEINTRESOURCE(IDD_SETTINGS_SCENESW),
                             parentHwnd, (DLGPROC)DialogProc, (LPARAM)this);

	//EM_SETCUEBANNER


	HWND hwndMainScn = GetDlgItem(hwnd, IDC_MAINSCN);
	HWND hwndTitle = GetDlgItem(hwnd, IDC_EDIT1);
	HWND hwndGames = GetDlgItem(hwnd, IDC_EDIT2);

	ComboBox_SetCueBannerText(hwndMainScn, PluginStr("Settings.Scene"));
	Edit_SetCueBannerText(hwndTitle, PluginStr("Settings.Title"));
	Edit_SetCueBannerText(hwndGames, PluginStr("Settings.GameTitle"));

    return hwnd;
}

void TwitchSwitcherSettings::DestroyPane()
{
    DestroyWindow(hwnd);
    hwnd = NULL;
}

INT_PTR TwitchSwitcherSettings::ProcMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message)
    {
    case WM_INITDIALOG:
        MsgInitDialog();
        return TRUE;
    case WM_COMMAND:
        return MsgClicked(LOWORD(wParam), HIWORD(wParam), (HWND)lParam);
    case WM_CTLCOLORSTATIC:
		if(GetWindowLong((HWND)lParam, GWL_ID) == IDC_RUN)
		{
			HDC hdc = (HDC)wParam;
			SetTextColor(hdc, thePlugin->IsRunning() ? RGB(0,255,0) : RGB(255,0,0));
			SetBkColor(hdc, GetSysColor(COLOR_3DFACE));
			return (INT_PTR)GetSysColorBrush(COLOR_3DFACE);
		}
        break;
    case WM_NOTIFY:
        switch (LOWORD(wParam)) {
        case IDC_WSMAP:
            {
		        const NMITEMACTIVATE* lpnmitem = (LPNMITEMACTIVATE)lParam;
		        if(lpnmitem->hdr.idFrom == IDC_WSMAP && lpnmitem->hdr.code == NM_CLICK)
		        {
			        const int sel = lpnmitem->iItem;
			        if(sel >= 0)
			        {
				        HWND wsMap = GetDlgItem(hwnd, IDC_WSMAP);
				        HWND hwndAppList = GetDlgItem(hwnd, IDC_APPLIST);
				        HWND hwndMainScn = GetDlgItem(hwnd, IDC_MAINSCN);

				        // Get the text from the item
				        String title;
				        title.SetLength(256);
						String game;
						game.SetLength(256);
				        ListView_GetItemText(wsMap, sel, 0, title, 256);
						ListView_GetItemText(wsMap, sel, 1, game, 256);
				        String scn;
				        scn.SetLength(256);
				        ListView_GetItemText(wsMap, sel, 2, scn, 256);

				        // Set the combos
				        SetWindowText(hwndAppList, title);
				        SendMessage(hwndMainScn, CB_SETCURSEL, SendMessage(hwndMainScn, CB_FINDSTRINGEXACT, -1, (LPARAM)scn.Array()), 0);
			        }
                }
                return TRUE;
            }
            break;
        }
    }
    return FALSE;
}

void TwitchSwitcherSettings::MsgInitDialog()
{
    LocalizeWindow(hwnd, thePlugin->pluginLocale);

    HWND hwndAppList = GetDlgItem(hwnd, IDC_APPLIST);
	HWND hwndMainScn = GetDlgItem(hwnd, IDC_MAINSCN);
	HWND hwndAltScn  = GetDlgItem(hwnd, IDC_ALTSCN);
	HWND hwndWSMap = GetDlgItem(hwnd, IDC_WSMAP);
	HWND hwndSwButton = GetDlgItem(hwnd, IDC_ALTSWITCH);
	HWND hwndNoswButton = GetDlgItem(hwnd, IDC_ALTNOSWITCH);
	HWND hwndCurrent = GetWindow(GetDesktopWindow(), GW_CHILD); // The top child of the desktop

	// let's fill the listcontrol
	SendMessage(hwndWSMap, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT | LVS_EX_LABELTIP, LVS_EX_FULLROWSELECT | LVS_EX_LABELTIP);
	LVCOLUMN col1;
	col1.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
	col1.fmt = LVCFMT_RIGHT | LVCFMT_FIXED_WIDTH;
	col1.cx = 205;
	col1.pszText = (LPWSTR)PluginStr("Settings.Title");
	LVCOLUMN col2;
	col2.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
	col2.fmt = LVCFMT_LEFT | LVCFMT_FIXED_WIDTH;
	col2.cx = 204;
	col2.pszText = (LPWSTR)PluginStr("Settings.GameTitle");
	LVCOLUMN col3;
	col3.mask = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH;
	col3.fmt = LVCFMT_LEFT | LVCFMT_FIXED_WIDTH;
	col3.cx = 205;
	col3.pszText = (LPWSTR)PluginStr("Settings.Scene");
	SendMessage(hwndWSMap, LVM_INSERTCOLUMN, 0, (LPARAM) (LPLVCOLUMN) &col1);
	SendMessage(hwndWSMap, LVM_INSERTCOLUMN, 1, (LPARAM) (LPLVCOLUMN) &col2);
	SendMessage(hwndWSMap, LVM_INSERTCOLUMN, 2, (LPARAM)(LPLVCOLUMN)&col3);
	// Add the items
	for (int i = 0; i < thePlugin->GetnWindowsDefined(); i++)
	{
		String title = thePlugin->GetTitle(i); 
		String game = thePlugin->GetGame(i);
		String scene = thePlugin->GetScene(i);
		LVITEM item;
		item.iItem = 0;
		item.iSubItem = 0;
		item.mask = LVIF_TEXT;
		item.pszText = title;

		LVITEM subItem;
		subItem.iItem = 0;
		subItem.iSubItem = 1;
		subItem.mask = LVIF_TEXT;
		subItem.pszText = game;

		LVITEM subItem2;
		subItem2.iItem = 0;
		subItem2.iSubItem = 2;
		subItem2.mask = LVIF_TEXT;
		subItem2.pszText = scene;

		SendMessage(hwndWSMap, LVM_INSERTITEM, 0, (LPARAM) &item);
		SendMessage(hwndWSMap, LVM_SETITEM, 0, (LPARAM) &subItem);
		SendMessage(hwndWSMap, LVM_SETITEM, 0, (LPARAM)&subItem2);
	}
	// Set the radio buttons and alt scene combo state
	SendMessage(hwndSwButton, BM_SETCHECK, (thePlugin->IsAltDoSwitch() ? BST_CHECKED : BST_UNCHECKED), 0);
	SendMessage(hwndNoswButton, BM_SETCHECK, (thePlugin->IsAltDoSwitch() ? BST_UNCHECKED : BST_CHECKED), 0);
	EnableWindow(hwndAltScn, thePlugin->IsAltDoSwitch());

	do
	{
		if(IsWindowVisible(hwndCurrent))
		{
			// Get the styles for the window
			DWORD exStyles = (DWORD)GetWindowLongPtr(hwndCurrent, GWL_EXSTYLE);
			DWORD styles = (DWORD)GetWindowLongPtr(hwndCurrent, GWL_STYLE);

			if( (exStyles & WS_EX_TOOLWINDOW) == 0 && (styles & WS_CHILD) == 0)
			{
				// The window is not a toolwindow, and not a child window
				String strWindowName;

				// Get the name of the window
				strWindowName.SetLength(GetWindowTextLength(hwndCurrent));
				GetWindowText(hwndCurrent, strWindowName, strWindowName.Length()+1);
				// Add the Name of the window to the appList
				const int id = (int)SendMessage(hwndAppList, CB_ADDSTRING, 0, (LPARAM)strWindowName.Array());
				// Set the data for the added item to be the window handle
				SendMessage(hwndAppList, CB_SETITEMDATA, id, (LPARAM)hwndCurrent);
			}
		}
	} // Move down the windows in z-order
	while (hwndCurrent = GetNextWindow(hwndCurrent, GW_HWNDNEXT));

	// Get the list of scenes from OBS
	XElement* scnList = OBSGetSceneListElement();
	if(scnList)
	{
		const DWORD numScn = scnList->NumElements();
		for(DWORD i=0; i<numScn; i++)
		{
			CTSTR sceneName = (scnList->GetElementByID(i))->GetName();
			// Add the scene name to both scene lists
			SendMessage(hwndMainScn, CB_ADDSTRING, 0, (LPARAM)sceneName);
		}
	}

	// Set the frequency from the settings
	SetDlgItemInt(hwnd, IDC_FREQ, thePlugin->GettimeToSleep(), FALSE);

	SetDlgItemText(hwnd, IDC_FREQ2, thePlugin->GetOAuthToken());
	SetDlgItemText(hwnd, IDC_FREQ3, thePlugin->GetManTitle());
	SetDlgItemText(hwnd, IDC_FREQ5, thePlugin->GetManGame());

	// Set the autostart checkbox value from the settings
	if(thePlugin->IsStartAuto())
		CheckDlgButton(hwnd, IDC_STARTAUTO, BST_CHECKED);

	// Set the toggle hotkey control
	SendMessage(GetDlgItem(hwnd, IDC_TOGGLEHOTKEY), HKM_SETHOTKEY, thePlugin->GetToggleHotkey(), 0);

	// If the plugin is running, update the text values
	if(thePlugin->IsRunning())
	{
		SetWindowText(GetDlgItem(hwnd, IDC_RUN), PluginStr("Settings.Running"));
		SetWindowText(GetDlgItem(hwnd, IDC_STOP), PluginStr("Settings.Stop"));
	}
}

std::string MBFromW2(LPCWSTR pwsz) {
	int cch = WideCharToMultiByte(CP_ACP, 0, pwsz, -1, 0, 0, NULL, NULL);

	char* psz = new char[cch];

	WideCharToMultiByte(CP_ACP, 0, pwsz, -1, psz, cch, NULL, NULL);

	std::string st(psz);
	delete[] psz;

	return st;
}

INT_PTR TwitchSwitcherSettings::MsgClicked(int controlId, int code, HWND controlHwnd)
{
    switch(controlId)
	{
	case IDC_CLEAR_HOTKEY:
		if(code == BN_CLICKED) {
			SendMessage(GetDlgItem(hwnd, IDC_TOGGLEHOTKEY), HKM_SETHOTKEY, 0, 0);
            SetChangedSettings(pChange = true);
            return TRUE;
        }
        break;

	case IDC_STOP:
		if(code == BN_CLICKED) // Stop button clicked
		{
			if(thePlugin->IsRunning())
				thePlugin->StopThread(hwnd);
			else
			{
				ApplyConfig(hwnd);
                pChange = false;
				thePlugin->StartThread(hwnd);
			}
            SetChangedSettings(pChange);
            return TRUE;
		}
        break;
		
	case IDADD:
		{
			HWND wsMap = GetDlgItem(hwnd, IDC_WSMAP);
			HWND titleList = GetDlgItem(hwnd, IDC_EDIT1);
			HWND gameList = GetDlgItem(hwnd, IDC_EDIT2);
			HWND scnList = GetDlgItem(hwnd, IDC_MAINSCN);
				
			String wnd = GetEditText(titleList);
			// First column
			LVITEM lv1;
			lv1.mask = LVIF_TEXT;
			lv1.iItem = 0;
			lv1.iSubItem = 0;
			lv1.pszText = wnd;
			// Second column
			String gme = GetEditText(gameList);
			LVITEM lv2;
			lv2.mask = LVIF_TEXT;
			lv2.iItem = 0;
			lv2.iSubItem = 1;
			lv2.pszText = gme;
			// third column
			String scn = GetCBText(scnList, CB_ERR);
			LVITEM lv3;
			lv3.mask = LVIF_TEXT;
			lv3.iItem = 0;
			lv3.iSubItem = 2;
			lv3.pszText = scn;
			// Add first column then set second
			SendMessage(wsMap, LVM_INSERTITEM, 0, (LPARAM)&lv1);
			SendMessage(wsMap, LVM_SETITEM, 0, (LPARAM)&lv2);
			SendMessage(wsMap, LVM_SETITEM, 1, (LPARAM)&lv3);
            SetChangedSettings(pChange = true);
            return TRUE;
		}
        break;

	case IDREM:
		{
			// Remove the item
			HWND wsMap = GetDlgItem(hwnd, IDC_WSMAP);
			const int sel = SendMessage(wsMap, LVM_GETSELECTIONMARK, 0, 0);
			if (sel > -1)
				SendMessage(wsMap, LVM_DELETEITEM, sel, 0);

            SetChangedSettings(pChange = true);
		    return TRUE;
        }
        break;

	case IDC_SET:
		if (code == BN_CLICKED) // Stop button clicked
		{
			extern std::string channeltitle;
			extern std::string gametitle;
			
			channeltitle = MBFromW2(GetEditText(GetDlgItem(hwnd, IDC_FREQ3)));
			gametitle = MBFromW2(GetEditText(GetDlgItem(hwnd, IDC_FREQ5)));

			
			ApplyConfig(hwnd);
			return TRUE;
		}
		break;
	case IDC_ORDER:
		if (code == BN_CLICKED) // Stop button clicked
		{
			//Log(L"In order");

			//ApplyConfig(hwnd);
			//extern String lastSceneName;
			//lastSceneName = "Order";
			//return TRUE;
		}
		break;
	case IDC_RANDOM:
		if (code == BN_CLICKED) // Stop button clicked
		{
			extern String lastSceneName;
			lastSceneName = "Random";
			//ApplyConfig(hwnd);
			return TRUE;
		}
		break;
	case IDC_CHECK1:
		if (code == BN_CLICKED) // Stop button clicked
		{
			if (SendMessage(GetDlgItem(hwnd, IDC_CHECK1), BM_GETCHECK, 0, 0) == BST_CHECKED)
			{
				EnableWindow(GetDlgItem(hwnd, IDC_FREQ3), true);
				EnableWindow(GetDlgItem(hwnd, IDC_FREQ5), true);
			}
			else
			{
				EnableWindow(GetDlgItem(hwnd, IDC_FREQ3), false);
				EnableWindow(GetDlgItem(hwnd, IDC_FREQ5), false);
			}
			//SetChangedSettings(true);
			return TRUE;
		}
		break;

	case IDC_ALTSWITCH:
	case IDC_ALTNOSWITCH:
		if (code == BN_CLICKED)
		{
			HWND swButton = GetDlgItem(hwnd, IDC_ALTSWITCH);
			HWND altCombo = GetDlgItem(hwnd, IDC_ALTSCN);
			const bool swChecked = (SendMessage(swButton, BM_GETSTATE, 0, 0) & BST_CHECKED) != 0;
			EnableWindow(altCombo, swChecked);
            pChange = pChange || (swChecked != thePlugin->IsAltDoSwitch());
            SetChangedSettings(pChange);
            return TRUE;
		}
		break;

	case IDC_STARTAUTO:
        {
            HWND control = GetDlgItem(hwnd, IDC_STARTAUTO);
            bool newState = (SendMessage(control, BM_GETSTATE, 0, 0) & BST_CHECKED) != 0;
            pChange = pChange || (newState != thePlugin->IsStartAuto());
            SetChangedSettings(pChange);
            return TRUE;
        }
        break;
    case IDC_TOGGLEHOTKEY:
        if (code == EN_CHANGE) {
            SetChangedSettings(pChange = true);
            return TRUE;
        }
        break;
    case IDC_FREQ:
        if(code == EN_CHANGE)
        {
            DWORD newFreq = GetDlgItemInt(hwnd, IDC_FREQ, NULL, FALSE);
            DWORD oldFreq = thePlugin->GettimeToSleep();
            pChange = pChange || (newFreq != oldFreq);
            SetChangedSettings(pChange);
            return TRUE;
        }
		break;
	case IDC_FREQ2:
		if (code == EN_CHANGE)
		{
			String newoauth = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
			GetDlgItemText(hwnd, IDC_FREQ2, newoauth, 50);
			pChange = pChange || newoauth != thePlugin->GetOAuthToken();
			SetChangedSettings(pChange);
			return TRUE;
		}
		break;
	case IDC_FREQ3:
		if (code == EN_CHANGE)
		{
			String newmantitle = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
			GetDlgItemText(hwnd, IDC_FREQ3, newmantitle, 50);
			pChange = pChange || newmantitle != thePlugin->GetManTitle();
			SetChangedSettings(pChange);
			return TRUE;
		}
		break;
	case IDC_FREQ5:
		if (code == EN_CHANGE)
		{
			String newgametitle = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
			GetDlgItemText(hwnd, IDC_FREQ5, newgametitle, 50);
			pChange = pChange || newgametitle != thePlugin->GetManGame();
			SetChangedSettings(pChange);
			return TRUE;
		}
		break;
	case IDC_EDIT1:
		if (code == EN_CHANGE)
		{
			TCHAR newFreq[80];
			SendMessage(GetDlgItem(hwnd, IDC_EDIT1), WM_GETTEXT, 79, (LPARAM)newFreq);
			//pChange = pChange || newFreq != thePlugin->GetOAuthToken();
			//SetChangedSettings(true);
			//std::cout << 
			//Log(newFreq);
			return TRUE;
		}
		break;
	case IDC_EDIT2:
		if (code == EN_CHANGE)
		{
			TCHAR newFreq[80];
			SendMessage(GetDlgItem(hwnd, IDC_EDIT1), WM_GETTEXT, 79, (LPARAM)newFreq);
			//pChange = pChange || newFreq != thePlugin->GetOAuthToken();
			//SetChangedSettings(true);
			//std::cout << 
			//Log(newFreq);
			return TRUE;
		}
        break;
        case IDC_APPLIST:
		case IDC_MAINSCN: /*
            if (code == CBN_SELCHANGE || code == CBN_EDITCHANGE) {
                EditItem(code == CBN_SELCHANGE && controlId == IDC_APPLIST);
                return TRUE;
            }
            break; */
        case IDC_ALTSCN:
            if (code == CBN_SELCHANGE) {
                SetChangedSettings(pChange = true);
                return TRUE;
            }
            break;
	}
    return FALSE;
}

void TwitchSwitcherSettings::EditItem(bool selChange) {
    HWND wsMap = GetDlgItem(hwnd, IDC_WSMAP);
	HWND titleList = GetDlgItem(hwnd, IDC_EDIT1);
	HWND gameList = GetDlgItem(hwnd, IDC_EDIT2);
	HWND scnList = GetDlgItem(hwnd, IDC_MAINSCN);
				
	const int sel = SendMessage(wsMap, LVM_GETSELECTIONMARK, 0, 0);
	if (sel < 0) return;

	String gme = selChange ? GetCBText(titleList, CB_ERR) : GetEditText(titleList);
	String tit = selChange ? GetCBText(gameList, CB_ERR) : GetEditText(gameList);
	// First column
	LVITEM lv1;
	lv1.mask = LVIF_TEXT;
	lv1.iItem = sel;
	lv1.iSubItem = 0;
	lv1.pszText = tit;
	// Second column
	LVITEM lv2;
	lv2.mask = LVIF_TEXT;
	lv2.iItem = sel;
	lv2.iSubItem = 1;
	lv2.pszText = gme;
	// third column
	String scn = GetCBText(scnList, CB_ERR);
	LVITEM lv3;
	lv3.mask = LVIF_TEXT;
	lv3.iItem = sel;
	lv3.iSubItem = 2;
	lv3.pszText = scn;

	// Set the text
	SendMessage(wsMap, LVM_SETITEM, 0, (LPARAM)&lv1);
	SendMessage(wsMap, LVM_SETITEM, 0, (LPARAM)&lv2);
	SendMessage(wsMap, LVM_SETITEM, 0, (LPARAM)&lv3);
    SetChangedSettings(pChange = true);
}

void TwitchSwitcherSettings::ApplyConfig(HWND hWnd) {
	thePlugin->ApplyConfig(hWnd);

	// Redefine the hotkey
	if (hotkeyID)
		OBSDeleteHotkey(hotkeyID);

	if (thePlugin->GetToggleHotkey() != 0)
		hotkeyID = OBSCreateHotkey((DWORD)thePlugin->GetToggleHotkey(), ToggleHotkeyProc, 0);
}

void TwitchSwitcherSettings::ApplySettings()
{
    pChange = false;

    ApplyConfig(hwnd);
}

void TwitchSwitcherSettings::CancelSettings()
{
    pChange = false;
}

bool TwitchSwitcherSettings::HasDefaults() const
{
    return false;
}